* TablMast.prg           (c) 1997-2000, Peisch Custom Software, Inc.
*   Builds tables and indexes from TablMast.dbf
*   Will update structures and build new tags as needed.
*   ONLY WORKS WITH FREE TABLES

* Will Return 0 if no errors, 1 if errors were encountered

* 07/27/00 - Modified for public consumption
* 07/22/97 - Original version

* This function needs the following components, which are included in the
* zip file:
*    NDXFORM.SCX/SCT  - for displaying progress while building indexes
*    DB_EXCL.PRG      - for getting exclusive use of tables
*    TABLMAST.DBF/FPT - Master table of structure and index tags

* ProgName is a program or module you can pass if you only want to update
* specific files.  Passing ALL will update all tables.  Anything else will
* only update files where the PrgName field in TABLEMAST.DBF matches what
* you've pass for ProgName.

* Pck is an optional parameter.  If you want this routine to pack the 
* tables, send .T. for Pck.  TABLE STRUCTURES WILL NOT BE UPDATED OR
* CHANGED IF YOU ARE PACKING!

* DataPath is an optional parameter if you want the tables created in
* a directory other than the current directory.  The directory will
* be created if it doesn't exist.  DataPath can be either a relative
* or absolute path.

* This function requires TABLMAST.DBF for structure and index
* information.  See the example records included in that table to 
* show how to build up the table.

* You can create tables with extensions other than DBF by specifying
* that extension in TABLMAST.DBF

* You may add any fields you wish to TABLMAST to enhance it's function.

* You may put comments in the NDXCommand field as long as the line starts
* with an asterisk (*).  You cannot put comments in the CrCommand field.

* EXAMPLE OF HOW TO CALL THIS PROGRAM:

*    nErrs = TABLMAST('ALL',.F.,'DATA')
*    if nErrs = 1
*       messagebox('Problems creating/updating tables')
*       quit
*    endif

* The above example will create/update all the tables in TABLMAST.DBF,
* will NOT pack them, and will create them in a subdirectory called DATA.

parameter ProgName, Pck, DataPath
local CrCmd, NdxCmd, Cmd, Loc, Loc2, TabFlds, CurFlds, NeedUpdate, I, J, CdxName, TName, TagPos
local OldMWid, ErrFlag, TabAlias

ErrFlag = .F.

do form NdxForm

if empty(ProgName)
	ProgName = ''
endif

if empty(DataPath)
	DataPath = '.\'
else
	if not directory(Datapath)
		md (DataPath)
	endif
endif
if right(DataPath,1) <> '\'
	DataPath = Datapath+'\'
endif

OldMWid = set('memowidth')
set memowidth to 100

* Dimension arrays to avoid a compilation error
dimension TabStru[1], CurStru[1]

select 0
use Tablmast
scan
	if PrgName = 'ALL' or upper(ProgName)==trim(PrgName)
		NeedUpdate = .F.
		TabName = trim(DataPath)+trim(TablName)
		Loc = rat('.', TabName)
		Loc2 = rat('\', TabName)
		if Loc = 0
			if Loc2 = 0
				TabAlias = TabName
			else
				TabAlias = right(TabName,len(TabName)-Loc2)
			endif
		else
			if Loc2 = 0
				TabAlias = left(TabName,Loc-1)
			else
				TabAlias = substr(TabName,Loc2+1,Loc-Loc2-1)
			endif
		endif
		if not Pck		&& Don't check for structure changes if we're packing
			CrCmd = Crcommand
			if not iif(at('.',TabName)=0,file(TabName+'.dbf'),file(TabName))
				create table (TabName) &CrCmd
				use
				select TablMast
			else
				* See if we need to change the structure.
				select 0
				use (TabName)
				TabFlds = afields('TabStru')
				create cursor Temp &CrCmd
				CurFlds = afields('CurStru')
			
				NeedUpdate = .F.
				if TabFlds <> CurFlds
					* The number of fields aren't even the same.  We know we need to update.
					NeedUpdate = .T.
				else
					* Compare the fields in the 2 arrays to see if everything is the same.
					for I = 1 to alen(TabStru,1)
						for J = 1 to 4
							if TabStru(I, J) <> CurStru(I, J)
								NeedUpdate = .T.
								exit
							endif
						endfor
					endfor
				endif
			endif
		endif
		
		if (NeedUpdate or Pck) and db_excl(TabName,TabAlias)
			if Pck
				NdxForm.ShouldShow = .T.
				NdxForm.Show()
				? 'Packing '+TabName+'...'
				set talk on
			endif
			select (TabAlias)
			pack
			set talk off
			use
			if not Pck
				select Temp
				append from (TabName)
				copy to (TabName)
			endif
		endif
		
		if used(TabAlias)
			use in (TabAlias)
		endif
			
		do case
		case NeedUpdate
			CdxName = TabName+'.cdx'
			if file(CdxName)
				delete file (CdxName)
			endif
		case Pck
			* Don't check indexes
			select TablMast
		otherwise
			if empty(TablMast.NdxCommand)
				select 0
				use (TabName)
				if not empty(tag(1))
					NeedUpdate = .T.
				endif
				use in (TabAlias)
			else
				* Find out if the tags are all the same.  If not, we need to re-create the index.
				dimension cTags[memlines(TablMast.NdxCommand),2]
				cTags = .F.
				* Fill the array with the tag names
				for I = 1 to memlines(TablMast.NdxCommand)
					if ltrim(mline(TablMast.NdxCommand, I)) = '*'
						loop
					endif
					* The tag name is to the right of the word TAG
					TName = upper(trim(mline(TablMast.NdxCommand, I)))
					if empty(TName)
						loop
					endif
					TName = right(TName, len(TName)-(at('TAG ',TName)+3))
					* If there's anything after the tag name (like CANDIDATE) remove it
					if at(' ',TName) <> 0
						TName = trim(left(TName, at(' ',TName)))
					endif
					cTags[I,1] = TName
				endfor
			
				* Look for all the tags we have and compare with what we need
				select 0
				use (TabName)
				I = 1
				do while .T.
					if empty(Tag(I))
						exit
					endif
					set exact on
					TagPos = ascan(cTags, Tag(I))
					set exact off
					if TagPos = 0
						NeedUpdate = .T.
						exit
					endif
					TagPos = asubscript(cTags, TagPos, 1)
					cTags[TagPos, 2] = .T.
					I = I+1
				enddo
				use in (TabAlias)

				* Look to see if we found tags for all the required tags
				for I = 1 to alen(Ctags,1)
					if not cTags[I, 2] and not empty(cTags[I, 1])
						NeedUpdate = .T.
						exit
					endif
				endfor
			endif
			
			if NeedUpdate
				CdxName = TabName+'.cdx'
				if file(CdxName)
					delete file (CdxName)
				endif
			endif
		endcase
		if used('Temp')
			use in Temp
		endif
		select TablMast

		if not empty(Ndxcommand) and not file(TabName+'.cdx')
			select 0
			use (TabName)
			if db_excl(TabName,TabAlias)
				NdxCmd = TablMast.Ndxcommand
				NdxForm.ShouldShow = .T.
				NdxForm.Show()
				? 'Building '+TabName+'.cdx...'
				do while len(NdxCmd) > 0
					* Separate each line in the memo field and execute it separately
					Loc = at(chr(13), NdxCmd)
					if Loc = 0
						Cmd = NdxCmd
					else
						Cmd = left(NdxCmd, Loc-1)
					endif
					NdxCmd = right(NdxCmd,len(NdxCmd)-(len(Cmd)+2))
					set talk on
					&Cmd
					set talk off
				enddo
			else
				ErrFlag = .T.
			endif
		endif
	
		if used(TabAlias)
			if not empty(TablMast.Ndxcommand)
				select (TabAlias)
				* If the table has an extension other than DBF, we can't use
				* the TabName.
				set index to (DataPath+TabAlias+'.cdx')
			endif
			use in (TabAlias)
		endif
	endif
endscan

use in TablMast

if Pck
	?
	wait 'Packing complete' timeout 3
endif

NdxForm.release()
set memowidth to (OldMWid)

return iif(ErrFlag,1,0)
*------------------------------------------------------------------------------------